#pragma once

#include <type_traits>

#include <dune/common/hybridutilities.hh>
#include <dune/functions/functionspacebases/subentitydofs.hh>

#include <amdis/LinearAlgebra.hpp>
#include <amdis/functions/Interpolate.hpp>
#include <amdis/gridfunctions/GridFunction.hpp>

namespace AMDiS {
namespace Impl {

/// Traverse all interior boundary DOFs and apply the functor f.
/// The Functor must have the signature `void(int, typename Basis::LocalView, typename Basis::GridView::Intersection)`
template <class Basis, class F>
void forEachInteriorBoundaryDOF(Basis const& basis, F&& f)
{
  auto localView = basis.localView();
  auto seDOFs = Dune::Functions::subEntityDOFs(basis);
  auto const& gridView = basis.gridView();
  for (auto const& element : elements(gridView, typename BackendTraits<Basis>::PartitionSet{})) {
    if (element.hasBoundaryIntersections()) {
      localView.bind(element);
      for(auto const& intersection: intersections(gridView, element))
        if (intersection.boundary())
          for(auto localIndex: seDOFs.bind(localView,intersection))
            f(localIndex, localView, intersection);
    }
  }
}

} // end namespace Impl


template <class D, class R>
  template <class RB, class CB>
void DirichletBC<D,R>::
init(RB const& rowBasis, CB const& colBasis)
{
  using LV = typename CB::LocalView;
  using IS = typename CB::GridView::Intersection;
  dirichletNodes_.resize(colBasis.dimension());
  dirichletNodes_.assign(dirichletNodes_.size(), false);
  Impl::forEachInteriorBoundaryDOF(colBasis, [&](int localIndex, LV const& localView, IS const& intersection) {
    dirichletNodes_[localView.index(localIndex)] = dirichletNodes_[localView.index(localIndex)] || onBoundary(intersection);
  });
}


template <class D, class R>
  template <class Mat, class Sol, class Rhs, class RN, class RTP, class CN, class CTP>
void DirichletBC<D,R>::
fillBoundaryCondition(Mat& matrix, Sol& solution, Rhs& rhs, RN const& rowNode, RTP rowTreePath, CN const& colNode, CTP colTreePath)
{
  Dune::Hybrid::ifElse(std::is_same<RangeType_t<RN>, R>{},
    [&](auto id) {
      auto&& gf = makeGridFunction(values_, solution.basis()->gridView());
      AMDiS::interpolate(*solution.basis(), id(solution), gf, rowTreePath, tag::defaulted{}, dirichletNodes_);
    });

  dirichletBC(matrix, solution, rhs, dirichletNodes_);
}

} // end namespace AMDiS
