#pragma once

#include <algorithm>
#include <list>
#include <memory>
#include <type_traits>

#include <dune/common/typeutilities.hh>

#include <amdis/common/ConceptsBase.hpp>
#include <amdis/Output.hpp>

namespace AMDiS
{
  namespace Impl
  {
    // forward declaration
    template <class Event>
    class ObserverInterface;

    template <class Event>
    class SignalBase
    {
    public:
      /// Attaches an observer to this class. This method will be called by all observers with
      /// with themselves as argument.
      void attachObserver(ObserverInterface<Event>* o) const
      {
        observers_.push_back(o);
      }

      /// Detaches an observer to this class. This method will be called by all observers with
      /// with themselves as argument.
      void detachObserver(ObserverInterface<Event>* o) const
      {
        auto it = std::find(observers_.begin(), observers_.end(), o);
        if (it != observers_.end())
          observers_.erase(it);
      }

      /// Notify all observers that have called attachObserver but have not called detachObserver
      void notify(Event const& e) const
      {
        for (ObserverInterface<Event>* o : observers_)
          o->update(e);
      }

    private:
      /// List of observers that need to be notified in case of an event
      // NOTE: this list is mutable, since the notification list itself is not part
      //       of the internal state of the object signaling the event.
      mutable std::list<ObserverInterface<Event>*> observers_;
    };


    template <class Event>
    class ObserverInterface
    {
    public:
      virtual ~ObserverInterface() = default;

      /// Attach the observer to a subject. It will then receive notifications from the subject when
      /// an event of type Event it triggered.
      void attach(std::shared_ptr<SignalBase<Event> const> const& subject)
      {
        if (bool(subject))
          subject->attachObserver(this);
      }

      /// Detach the observer from the subject. It will then no longer receive notifications. This
      /// must be called before the observer is deleted.
      void detach(std::shared_ptr<SignalBase<Event> const> const& subject)
      {
        if (bool(subject))
          subject->detachObserver(this);
      }

      /// This method will be called by the subject when triggering the event.
      virtual void update(Event const&)
      {
        error_exit("Method must be overridden by derived class");
      }
    };


    template <class Event>
    class ObserverBase
        : virtual public ObserverInterface<Event>
    {
      using Self = ObserverBase;

    private:
      ObserverBase() = default;

      // use subject if Event is handled directly
      template <class S,
        REQUIRES(std::is_base_of<SignalBase<Event>, S>::value)>
      ObserverBase(std::shared_ptr<S const> subject, Dune::PriorityTag<2>)
        : subject_(std::move(subject))
      {
        this->attach(subject_);
      }

      // get next subject in hierarchy if Event is not handled
      template <class Observer,
        class = void_t<decltype(std::declval<Observer>().subject_)> >
      ObserverBase(std::shared_ptr<Observer const> const& o, Dune::PriorityTag<1>)
        : ObserverBase(o->subject_)
      {}

      // non-observable type
      template <class T>
      ObserverBase(std::shared_ptr<T const> const& other, Dune::PriorityTag<0>)
      {
        /* fallback implementation */
      }

    public:
      template <class T>
      ObserverBase(std::shared_ptr<T const> const& arg)
        : ObserverBase(arg, Dune::PriorityTag<42>{})
      {}

      /// Copy constructor
      ObserverBase(Self const& that)
        : subject_(that.subject_)
      {
        this->attach(subject_);
      }

      /// Move constructor
      ObserverBase(Self&& that)
      {
        swap(*this, that);
      }

      /// Destructor. Detaches observer from the subject
      ~ObserverBase() override
      {
        this->detach(subject_);
      }

      /// Copy/Move assignment
      Self& operator=(Self that)
      {
        swap(*this, that);
        return *this;
      }

      /// Swaps the contents of lhs and rhs
      friend void swap(Self& lhs, Self& rhs)
      {
        using std::swap;
        lhs.detach(lhs.subject_);
        rhs.detach(rhs.subject_);
        swap(lhs.subject_, rhs.subject_);
        lhs.attach(lhs.subject_);
        rhs.attach(rhs.subject_);
      }

    private:
      /// The observed subject
      std::shared_ptr<SignalBase<Event> const> subject_ = nullptr;
    };

  } // end namespace Impl


  /// \brief Mixin for signaling of certain events.
  /**
   *  Derived classes T can emit a signal e by calling notify(e). This will send the signal to all
   *  classes U... deriving from Observer<S, Events...> if
   *  - the type of the event is included in Events,
   *  - T = S or S is an Observer of T (directly or indirectly via other Observers),
   *  - U called the Observer constructor with an instance of S that has direct or indirect access
   *      to the instance of T (see Observer)
   */
  template <class Event, class... Events>
  class Signals
      : public Impl::SignalBase<Event>
      , public Signals<Events...>
  {
  public:
    using Impl::SignalBase<Event>::notify;
    using Signals<Events...>::notify;
  };

  template <class Event>
  class Signals<Event>
      : public Impl::SignalBase<Event>
  {
  public:
    using Impl::SignalBase<Event>::notify;
  };


  /// \brief Mixin for reacting to certain events.
  /**
   *  Derived classes can react to events by implementing the functions `update(Event)` for
   *  all Events specified in the template parameter list, whose origin is an instance of class
   *  `Subject` specified in the constructor to Observer and any indirectly observed class instances.
   *  Observed instances may include:
   *  - the instance of class `Subject` passed to the constructor
   *  - the instance passed to the constructor of `Subject::Observer<S>` if Subject inherits from
   *    \ref Observer<S>
   *  - all other instances indirectly accessible in the above way
   *
   *  For each event E the first object in the above hierarchy that implements \ref Signals<Es...> with E
   *  included in Es... will be observed by this class.
   */
  template <class Subject, class... Events>
  class Observer
      : public Impl::ObserverBase<Events>...
  {
    template <class E>
    friend class Impl::ObserverBase;

  public:
    Observer(std::shared_ptr<Subject const> const& s)
      : Impl::ObserverBase<Events>(s)...
      , subject_(std::move(s))
    {}

    Observer(std::shared_ptr<Subject> const& s)
      : Observer(std::const_pointer_cast<Subject const>(s))
    {}

  private:
    std::shared_ptr<Subject const> subject_ = nullptr;
  };

} // end namespace AMDiS
