#pragma once

#include <functional>
#include <list>
#include <map>
#include <type_traits>
#include <vector>

#include <amdis/Output.hpp>
#include <amdis/common/ConceptsBase.hpp>
#include <amdis/common/FieldMatVec.hpp>


namespace AMDiS
{
  template <class ct, int dow>
  struct FaceTransformation
  {
    using WorldMatrix = FieldMatrix<ct, dow, dow>;
    using WorldVector = FieldVector<ct, dow>;

    WorldMatrix matrix_; // Note: We assume the matrix to be orthogonal, here
    WorldVector shift_;

    WorldVector evaluate(WorldVector const& x) const
    {
      WorldVector y = shift_;
      matrix_.umv(x, y);
      return y;
    }

    WorldVector evaluateInverse(WorldVector const& y) const
    {
      WorldVector ys = y - shift_;
      WorldVector x;
      matrix_.mtv(ys, x);
      return x;
    }
  };


  /// Implements a periodic boundary condition
  /**
   * \tparam D  Domain
   * \tparam MI Type of the MultiIndex
   **/
  template <class D, class MI>
  class PeriodicBC
      : public BoundaryCondition
  {
    using Super = BoundaryCondition;

  public:
    using Domain = D;
    using MultiIndex = MI;
    using FaceTrafo = FaceTransformation<typename D::field_type, D::dimension>;

  public:
    template <class BM>
    PeriodicBC(BM&& boundaryManager, BoundaryType id, FaceTrafo faceTrafo)
      : Super(FWD(boundaryManager), id)
      , faceTrafo_(std::move(faceTrafo))
    {}

    template <class RB, class CB>
    void init(RB const& rowBasis, CB const& colBasis);

    template <class Mat, class Sol, class Rhs, class RN, class RTP, class CN, class CTP>
    void fillBoundaryCondition(Mat& A, Sol& x, Rhs& b, RN const& rowNode, RTP rowTreePath, CN const& colNode, CTP colTreePath);

    auto const& associations() const
    {
      return associations_;
    }

    auto const& periodicNodes() const
    {
      return periodicNodes_;
    }

  protected:
    template <class Basis>
    void initAssociations(Basis const& basis);

    template <class Basis>
    void initAssociations2(Basis const& basis);

    // Get the coordinates of the DOFs
    template <class Node>
    std::vector<Domain> coords(Node const& node, std::vector<std::size_t> const& localIndices) const;

  private:
    FaceTrafo faceTrafo_;

    std::vector<bool> periodicNodes_;
    std::map<std::size_t, std::size_t> associations_;
    Dune::Std::optional<bool> hasConnectedIntersections_;
  };


  template <class Basis>
  struct PeriodicData
  {
    using WorldVector = typename Basis::GridView::template Codim<0>::Geometry::GlobalCoordinate;

    template <class RowNode, class ColNode>
    using type = std::list< std::shared_ptr<PeriodicBC<WorldVector, typename Basis::MultiIndex>> >;
  };

  template <class RowBasis, class ColBasis>
  using PeriodicBCs = MatrixData<RowBasis, ColBasis, PeriodicData<RowBasis>::template type>;

} // end namespace AMDiS

#include "PeriodicBC.inc.hpp"
