#pragma once

#include <string>

#include <amdis/ProblemInstatBase.hpp>
#include <amdis/ProblemStat.hpp>
#include <amdis/common/TypeTraits.hpp>

namespace AMDiS
{
  // forward declarations
  class AdaptInfo;

  /**
   * \ingroup Problem
   *
   * \brief
   * Standard implementation of ProblemTimeInterface for a time
   * dependent problems.
   */
  template <class Traits>
  class ProblemInstat
      : public ProblemInstatBase
  {
    using Self        = ProblemInstat;
    using ProblemType = ProblemStat<Traits>;

    using CoefficientVector = typename ProblemType::CoefficientVector;

  public:
    /// Constructs a ProblemInstat with prob as its stationary problem, stored as reference.
    ProblemInstat(std::string const& name, ProblemType& prob)
      : ProblemInstatBase(name)
      , problemStat_(&prob)
    {}

    /// Constructor. Stores a reference to prob and initialProb.
    ProblemInstat(std::string const& name, ProblemType& prob, ProblemStatBase& initialProb)
      : ProblemInstatBase(name, initialProb)
      , problemStat_(&prob)
    {}

    /// Initialisation of the problem.
    void initialize(Flag initFlag = INIT_NOTHING);

    /// Implementation of \ref ProblemTimeInterface::initTimestep().
    void initTimestep(AdaptInfo& adaptInfo) override;

    /// Implementation of \ref ProblemTimeInterface::closeTimestep().
    void closeTimestep(AdaptInfo& adaptInfo) override;

    /// Returns \ref problemStat.
    ProblemType&       problemStat()       { return *problemStat_; }
    ProblemType const& problemStat() const { return *problemStat_; }

    /// Returns const-ref of \ref oldSolution.
    std::shared_ptr<CoefficientVector const> oldSolutionVector() const
    {
      test_exit_dbg(bool(oldSolution_),
        "OldSolution need to be created. Call initialize with INIT_UH_OLD.");
      return oldSolution_;
    }

    /// Return a const view to a oldSolution component
    template <class TreePath = RootTreePath>
    auto oldSolution(TreePath path = {}) const
    {
      test_exit_dbg(bool(oldSolution_),
        "OldSolution need to be created. Call initialize with INIT_UH_OLD.");
      return oldSolution_->child(path);
    }

    /// Implementation of \ref ProblemTimeInterface::transferInitialSolution().
    void transferInitialSolution(AdaptInfo& adaptInfo) override;

  protected:
    /// Used in \ref initialize() to create the \ref oldSolution_.
    void createUhOld();

  protected:
    /// Space problem solved in each timestep. (non-owning pointer)
    ProblemType* problemStat_;

    /// Solution of the last timestep.
    std::shared_ptr<CoefficientVector> oldSolution_;
  };


#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  // Deduction rule
  template <class Traits>
  ProblemInstat(std::string const& name, ProblemStat<Traits>& prob)
    -> ProblemInstat<Traits>;

  template <class Traits>
  ProblemInstat(std::string const& name, ProblemStat<Traits>& prob, ProblemStatBase& initialProb)
    -> ProblemInstat<Traits>;
#endif

  // Generator for ProblemInstat with given ProblemStat
  template <class Traits>
  ProblemInstat<Traits> makeProblemInstat(std::string const& name, ProblemStat<Traits>& prob)
  {
    return {name, prob};
  }

  // Generator for ProblemInstat with given ProblemStat and initialization problem
  template <class Traits>
  ProblemInstat<Traits> makeProblemInstat(std::string const& name, ProblemStat<Traits>& prob, ProblemStatBase& initialProb)
  {
    return {name, prob, initialProb};
  }

  // mark template as explicitly instantiated in cpp file
  extern template class ProblemInstat<YaspGridBasis<2,1>>;
  extern template class ProblemInstat<YaspGridBasis<2,2>>;

} // end namespace AMDiS

#include "ProblemInstat.inc.hpp"
