#pragma once

#include <dune/common/typeutilities.hh>

#include <amdis/common/ConceptsBase.hpp>
#include <amdis/common/Logical.hpp>

namespace AMDiS
{
  namespace Concepts
  {
    namespace Definition
    {
      struct VectorResizable
      {
        template <class V>
        auto require(V const& vec, Dune::PriorityTag<2>) -> decltype( const_cast<V&>(vec).resize(0u), 0);

        template <class V>
        auto require(V const& vec, Dune::PriorityTag<1>) -> decltype( const_cast<V&>(vec).change_dim(0u), 0);
      };

      struct MatrixResizable
      {
        template <class M>
        auto require(M const& mat, Dune::PriorityTag<3>) -> decltype( const_cast<M&>(mat).resize(0u,0u), 0);

        template <class M>
        auto require(M const& mat, Dune::PriorityTag<2>) -> decltype( const_cast<M&>(mat).change_dim(0u,0u), 0);

        template <class M>
        auto require(M const& mat, Dune::PriorityTag<1>) -> decltype( const_cast<M&>(mat).setSize(0u,0u), 0);
      };
    }

    /// Checks whether a vector can be resized by various resize methods
    template <class Vector>
    using VectorResizable_t = models_t<Definition::VectorResizable(Vector, Dune::PriorityTag<42>)>;

    /// Checks whether a matrix can be resized by various resize methods
    template <class Matrix>
    using MatrixResizable_t = models_t<Definition::MatrixResizable(Matrix, Dune::PriorityTag<42>)>;

  } // end namespace Concepts

#ifdef DOXYGEN
  /// \brief Uniform vector resize, using either vector.resize() or vector.change_dim()
  template <class Vector>
  void resize(Vector& vec, std::size_t size);

  /// \brief Uniform matrix resize, using either matrix.resize(), matrix.setSize() or matrix.change_dim()
  template <class Matrix>
  void resize(Matrix& mat, std::size_t r, std::size_t c);
#else

  namespace Impl
  {
    template <class Vector,
      class = void_t<decltype(std::declval<Vector&>().resize(0u))> >
    void resize_impl(Vector& vec, std::size_t size, Dune::PriorityTag<2>)
    {
      vec.resize(size);
    }

    template <class Vector,
      class = void_t<decltype(std::declval<Vector&>().change_dim(0u))> >
    void resize_impl(Vector& vec, std::size_t size, Dune::PriorityTag<1>)
    {
      vec.change_dim(size);
    }

    template <class Vector>
    void resize_impl(Vector& /*vec*/, std::size_t /*size*/, Dune::PriorityTag<0>)
    {
      /* do nothing */
    }
  }

  template <class Vector>
  void resize(Vector& vec, std::size_t size)
  {
    Impl::resize_impl(vec, size, Dune::PriorityTag<42>{});
  }


  namespace Impl
  {
    template <class Matrix,
      class = void_t<decltype(std::declval<Matrix&>().resize(0u,0u))> >
    void resize_impl(Matrix& mat, std::size_t rows, std::size_t cols, Dune::PriorityTag<3>)
    {
      mat.resize(rows, cols);
    }

    template <class Matrix,
      class = void_t<decltype(std::declval<Matrix&>().change_dim(0u,0u))> >
    void resize_impl(Matrix& mat, std::size_t rows, std::size_t cols, Dune::PriorityTag<2>)
    {
      mat.change_dim(rows, cols);
    }

    template <class Matrix,
      class = void_t<decltype(std::declval<Matrix&>().setSize(0u,0u))> >
    void resize_impl(Matrix& mat, std::size_t rows, std::size_t cols, Dune::PriorityTag<1>)
    {
      mat.setSize(rows, cols);
    }

    template <class Matrix>
    void resize_impl(Matrix& /*mat*/, std::size_t /*rows*/, std::size_t /*cols*/, Dune::PriorityTag<0>)
    {
      /* do nothing */
    }
  }

  template <class Matrix>
  void resize(Matrix& mat, std::size_t rows, std::size_t cols)
  {
    Impl::resize_impl(mat, rows, cols, Dune::PriorityTag<42>{});
  }

#endif

} // end namespace AMDiS
