#pragma once

/** \file LinearSolverInterface.h */

/**
 * \defgroup Solver Solver module
 * @{ <img src="solver.png"> @}
 *
 * \brief
 * Contains all classes needed for solving linear and non linear equation
 * systems.
 */

namespace AMDiS
{
  class SolverInfo;

  /// Abstract base class for linear solvers
  template <class Traits>
  class LinearSolverInterface
  {
  protected:
    using Mat = typename Traits::Mat;
    using Vec = typename Traits::Vec;
    using Comm = typename Traits::Comm;

  public:
    /// Destructor.
    virtual ~LinearSolverInterface() = default;

    /// Public method to call in order to solve a linear system Ax = b.
    /**
     * The method redirects to the specific linear solver and prints statistics
     * and error estimations at the end.
     *
     * The parameters correspond to
     *  \p A     A [block-]matrix that represents the system-matrix.
     *  \p x     A [block-]vector for the unknown components.
     *  \p b     A [block-]vector for the right-hand side of the linear system.
     **/
    void solve(Mat const& A, Vec& x, Vec const& b, Comm& comm, SolverInfo& solverInfo)
    {
      solveImpl(A, x, b, comm, solverInfo);
    }

  private:
    /// main methods that all solvers must implement
    virtual void solveImpl(Mat const& A, Vec& x, Vec const& b, Comm& comm, SolverInfo& solverInfo) = 0;
  };

} // end namespace AMDiS
