#pragma once

#include <string>

#include <amdis/Initfile.hpp>

namespace AMDiS
{
  /// Class that stores information about the solution process, like tolerances
  /// and iteration counts and is passed to the solver and filled there.
  class SolverInfo
  {
  public:
    /// The constructor reads needed parameters and sets solver \p prefix.
    /**
     * Reads parameters for a solver with name 'NAME':
     *   NAME->absolute tolerance   \ref aTol_
     *   NAME->relative tolerance   \ref rTol_
     *   NAME->info                 \ref info_
     *   NAME->break if tolerance not reached \ref breakTolNotReached_
    **/
    explicit SolverInfo(std::string const& prefix)
      : prefix_(prefix)
    {
      Parameters::get(prefix + "->absolute tolerance", aTol_);
      Parameters::get(prefix + "->relative tolerance", rTol_);
      Parameters::get(prefix + "->info", info_);
      Parameters::get(prefix + "->break if tolerance not reached", breakTolNotReached_);
    }

    /** \name getting methods
     * \{
     */

    /// Returns \ref aTol_
    double absTolerance() const
    {
      return aTol_;
    }

    /// Returns \ref rTol_
    double relTolerance() const
    {
      return rTol_;
    }

    /// Returns error code in last run of an iterative solver
    int errorCode() const
    {
      return error_;
    }

    /// Returns info
    int info() const
    {
      return info_;
    }

    /// Returns \ref absResidual_
    double absResidual() const
    {
      return absResidual_;
    }

    /// Returns \ref relResidual_
    double relResidual() const
    {
      return relResidual_;
    }

    /// Returns the initfile \ref prefix_
    std::string const& prefix() const
    {
      return prefix_;
    }

    /// Returns \ref createMatrixData
    bool doCreateMatrixData() const
    {
      return createMatrixData_;
    }

    /// Returns \ref storeMatrixData
    bool doStoreMatrixData() const
    {
      return storeMatrixData_;
    }

    bool doBreak() const
    {
      return breakTolNotReached_ && (
          (aTol_ > 1.e-30 && absResidual_ > aTol_) ||
          (rTol_ > 1.e-30 && relResidual_ > rTol_) );
    }

    /** \} */


    /** \name setting methods
     * \{
     */

    /// Sets \ref aTol_
    void setAbsTolerance(double tol)
    {
      aTol_ = tol;
    }

    /// Sets \ref rTol_
    void setRelTolerance(double tol)
    {
      rTol_ = tol;
    }

    /// Sets \ref aTol_
    void setAbsResidual(double r)
    {
      absResidual_ = r;
    }

    /// Sets \ref rTol_
    void setRelResidual(double r)
    {
      relResidual_ = r;
    }

    /// Sets \ref info_
    void setInfo(int i)
    {
      info_ = i;
    }

    /// Sets \ref error_
    void setError(int e)
    {
      error_ = e;
    }

    /// Sets \ref createMatrixData_
    void setCreateMatrixData(bool b)
    {
      createMatrixData_ = b;
    }

    /// Sets \ref storeMatrixData_
    void setStoreMatrixData(bool b)
    {
      storeMatrixData_ = b;
    }

    /** \} */

  private:
    /// The initfile prefix to read parameters
    std::string prefix_;

    /// The abolute tolerance
    double aTol_ = 0;

    /// The relative tolerance
    double rTol_ = 1.e-6;

    /// Throw an error if tolerance could not be reached
    bool breakTolNotReached_ = false;

    /// The solver verbosity level
    int info_ = 0;

    /// The absolute residual, default (-1): not set
    double absResidual_ = -1.0;

    /// The relative residual, relative to the two_norm(b), default (-1): not set
    double relResidual_ = -1.0;

    /// The error-code, default (-1): not set
    int error_ = -1;

    /// If true, the matrix will be initialized and the
    /// corresponding runner of the system receives the
    /// matrix in the init() method.
    bool createMatrixData_ = true;

    /// If false, the exit() method of the runner will be called.
    bool storeMatrixData_ = false;
  };

} // end namespace AMDiS
