#pragma once

#include <algorithm>
#include <string>

#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/linearalgebra/eigen/SolverConfig.hpp>

namespace AMDiS
{
  /**
  * \ingroup Solver
  * \class AMDiS::DirectRunner
  * \brief \implements RunnerInterface for the (external) direct solvers
  */
  template <template <class> class Solver, class Traits>
  class DirectRunner
      : public RunnerInterface<Traits>
  {
  protected:
    using Super = RunnerInterface<Traits>;
    using Mat = typename Traits::Mat;
    using Vec = typename Traits::Vec;
    using Comm = typename Traits::Comm;
    using EigenSolver = Solver<Mat>;

  public:
    /// Constructor.
    DirectRunner(std::string const& prefix)
      : solver_{}
    {
      SolverConfig<EigenSolver>::init(prefix, solver_);
      Parameters::get(prefix + "->reuse pattern", reusePattern_);
    }

    /// Implements \ref RunnerInterface::init()
    void init(Mat const& A, Comm& comm) override
    {
      DUNE_UNUSED_PARAMETER(comm);

      if (!reusePattern_ || !initialized_) {
        solver_.analyzePattern(A);
        initialized_ = true;
      }
      solver_.factorize(A);

      test_exit(solver_.info() == Eigen::Success,
        "Error in solver.compute(matrix)");
    }


    /// Implements \ref RunnerInterface::exit()
    void exit() override {}

    /// Implements \ref RunnerInterface::solve()
    int solve(Mat const& A, Vec& x, Vec const& b, SolverInfo& solverInfo) override
    {
      DUNE_UNUSED_PARAMETER(A);

      x = solver_.solve(b);

      auto r = Vec(b);
      if (x.norm() != 0)
        r -= A * x;

      solverInfo.setAbsResidual(r.norm());
      solverInfo.setError(solver_.info());

      return solver_.info() == Eigen::Success ? 0 : 1;
    }

  private:
    EigenSolver solver_;
    bool reusePattern_ = false;
    bool initialized_ = false;
  };
}
