#pragma once

#include <memory>

#include <dune/common/unused.hh>
#include <dune/common/version.hh>

#include <dune/istl/novlpschwarz.hh>
#include <dune/istl/schwarz.hh>
#include <dune/istl/paamg/amg.hh>
#include <dune/istl/paamg/fastamg.hh>
#include <dune/istl/paamg/kamg.hh>

#include <amdis/Initfile.hpp>
#include <amdis/common/ConceptsBase.hpp>
#include <amdis/linearalgebra/istl/Communication.hpp>
#include <amdis/linearalgebra/istl/CreatorInterfaces.hpp>
#include <amdis/linearalgebra/istl/precompiled/Preconditioners.hpp>
#include <amdis/linearalgebra/istl/precompiled/Solvers.hpp>

namespace AMDiS
{
  /// Create the final AMG object from passed \ref Dune::LinearOperator, a
  /// \ref CoarsenCriterion and \ref Dune::Amg::SmootherTraits arguments, with
  /// a parametrized Smoother type.
  template <template <class...> class AMGSolver, class Traits>
  struct AMGCreator;

  template <class Smoother>
  using SmootherArgs = typename Dune::Amg::SmootherTraits<Smoother>::Arguments;


  // Specialization for \ref Dune::Amg::AMG preconditioner
  template <class Traits>
  struct AMGCreator<Dune::Amg::AMG, Traits>
  {
    using PrecBase = typename Traits::Prec;

    template <class Smoother, class LinOp, class Criterion, class Communication>
    static std::unique_ptr<PrecBase>
    create(std::string prefix, LinOp const& linOp, Criterion const& criterion, SmootherArgs<Smoother> const& smootherArgs, Communication const& comm)
    {
      DUNE_UNUSED_PARAMETER(prefix);
      using Solver = Dune::Amg::AMG<LinOp, typename Traits::Vec, Smoother, Communication>;
      return std::make_unique<Solver>(linOp, criterion, smootherArgs, comm);
    }
  };


  // Specialization for \ref Dune::Amg::FastAMG preconditioner
  /**
   * Additional preconditioner initfile parameters:
   * - `[PRECON]->symmetric`:  Specifies whether the matrix is symmetric. [true]
   */
  template <class Traits>
  struct AMGCreator<Dune::Amg::FastAMG, Traits>
  {
    using PrecBase = typename Traits::Prec;

    template <class Smoother, class LinOp, class Criterion>
    static std::unique_ptr<PrecBase>
    create(std::string prefix, LinOp const& linOp, Criterion const& criterion, SmootherArgs<Smoother> const& smootherArgs,
           Dune::Amg::SequentialInformation const& comm)
    {
      DUNE_UNUSED_PARAMETER(smootherArgs);
      bool symmetric = Parameters::get<bool>(prefix + "->symmetric").value_or(true);

      using Solver = Dune::Amg::FastAMG<LinOp, typename Traits::Vec, Dune::Amg::SequentialInformation>;
      return std::make_unique<Solver>(linOp, criterion, criterion, symmetric, comm);
    }

    template <class Smoother, class LinOp, class Criterion, class SmootherArgs, class Communication,
      REQUIRES(not std::is_same<Communication, Dune::Amg::SequentialInformation>::value)>
    static std::unique_ptr<PrecBase>
    create(std::string, LinOp const&, Criterion const&, SmootherArgs const&, Communication const&)
    {
      error_exit("Currently only sequential FastAMG supported.");
      return nullptr;
    }
  };


  // Specialization for \ref Dune::Amg::KAMG preconditioner
  /**
   * Additional preconditioner initfile parameters:
   * - `[PRECON]->krylov solver`:  The key for the krylov solver used in the cycle. [pcg]
   * - `[PRECON]->krylov solver->maxLevelKrylovSteps`:  The number of krylov iterations. [3]
   * - `[PRECON]->krylov solver->minDefectReduction`:   The relative krylov solver tolerance. [1.e-1]
   *
   * Note: a flexible krylov solver should be used, like FCG or FGMRes. Possible values for the
   * krylov solver parameter: pcg, fcg, cfcg, [bicgstab|bcgs], minres.
   */
  template <class Traits>
  struct AMGCreator<Dune::Amg::KAMG, Traits>
  {
    using Vec = typename Traits::Vec;
    using PrecBase = typename Traits::Prec;

    template <class Smoother, class LinOp, class Criterion, class Communication>
    static std::unique_ptr<PrecBase>
    create(std::string prefix, LinOp const& linOp, Criterion const& criterion, SmootherArgs<Smoother> const& smootherArgs, Communication const& comm)
    {
      DUNE_UNUSED_PARAMETER(smootherArgs);
      std::string solver = Parameters::get<std::string>(prefix + "->krylov solver").value_or("default");

      std::size_t maxLevelKrylovSteps = 3;
      Parameters::get(prefix + "->krylov solver->maxLevelKrylovSteps", maxLevelKrylovSteps);
      double minDefectReduction = 1.e-1;
      Parameters::get(prefix + "->krylov solver->minDefectReduction", minDefectReduction);

      if (solver == "pcg" ||
          solver == "default")
      {
        using Solver = Dune::Amg::KAMG<LinOp, Vec, Smoother, Communication, Dune::GeneralizedPCGSolver<Vec>>;
        return std::make_unique<Solver>(linOp, criterion, smootherArgs, maxLevelKrylovSteps, minDefectReduction, comm);
      }
#if DUNE_VERSION_GTE(DUNE_ISTL,2,7)
      else if (solver == "fcg")
      {
        using Solver = Dune::Amg::KAMG<LinOp, Vec, Smoother, Communication, Dune::RestartedFCGSolver<Vec>>;
        return std::make_unique<Solver>(linOp, criterion, smootherArgs, maxLevelKrylovSteps, minDefectReduction, comm);
      }
      else if (solver == "cfcg")
      {
        using Solver = Dune::Amg::KAMG<LinOp, Vec, Smoother, Communication, Dune::CompleteFCGSolver<Vec>>;
        return std::make_unique<Solver>(linOp, criterion, smootherArgs, maxLevelKrylovSteps, minDefectReduction, comm);
      }
#endif
      else if (solver == "bicgstab" ||
               solver == "bcgs")
      {
        using Solver = Dune::Amg::KAMG<LinOp, Vec, Smoother, Communication, Dune::BiCGSTABSolver<Vec>>;
        return std::make_unique<Solver>(linOp, criterion, smootherArgs, maxLevelKrylovSteps, minDefectReduction, comm);
      }
      else if (solver == "minres")
      {
        using Solver = Dune::Amg::KAMG<LinOp, Vec, Smoother, Communication, Dune::MINRESSolver<Vec>>;
        return std::make_unique<Solver>(linOp, criterion, smootherArgs, maxLevelKrylovSteps, minDefectReduction, comm);
      }
#if 0
      // NOTE: can not be constructed inside the KAMG precon, since additional constructor argument.
      // Needs something like ConstructionTraits for solvers.
      else if (solver == "gmres")
      {
        using Solver = Dune::Amg::KAMG<LinOp, Vec, Smoother, Communication, Dune::RestartedGMResSolver<Vec>>;
        return std::make_unique<Solver>(linOp, criterion, smootherArgs, maxLevelKrylovSteps, minDefectReduction, comm);
      }
#endif
      else
      {
        error_exit("Unknown coarse space solver {} given for parameter `{}`.", solver, prefix + "->coarse space solver");
        return nullptr;
      }
    }
  };


  template <class Creator, class Smoother, class Traits>
  class AMGPrecon;

  /// A creator for \ref AMGPrecon, reads the smoother type from initfile:
  /**
   * Initfile parameters:
   * - `[PRECON]->smoother:  The Smoother type to use. [jacobi]
   *
   * Note: possible values for the smoother types: [diag|jacobi], ssor.
   */
  template <template <class...> class AMGSolver, class Traits>
  class AMGPreconCreator
      : public CreatorInterfaceName<ISTLPreconCreatorInterface<Traits>>
  {
    using Mat = typename Traits::Mat;
    using Vec = typename Traits::Vec;

    using PreconCreatorBase = ISTLPreconCreatorInterface<Traits>;
    using Creator = AMGCreator<AMGSolver,Traits>;

    template <class Smoother>
    using Precon = AMGPrecon<Creator, Smoother, Traits>;

  public:
    std::unique_ptr<PreconCreatorBase> createWithString(std::string prefix) override
    {
      // get creator string for preconditioner
      std::string smoother = "default";
      Parameters::get(prefix + "->smoother", smoother);

      if (smoother == "diag" ||
          smoother == "jacobi" ||
          smoother == "default")
      {
        auto creator = typename Precon<Dune::SeqJac<Mat, Vec, Vec>>::Creator{};
        return creator.createWithString(prefix);
      }
#if 0
      // NOTE: apply<forward> not implemented correctly in BlockPreconditioner and
      // NonoverlappingBlockPreconditioner. See !302 in dune-istl
      else if (smoother == "sor")
      {
        auto creator = typename Precon<Dune::SeqSOR<Mat, Vec, Vec>>::Creator{};
        return creator.createWithString(prefix);
      }
#endif
#if 0
      // NOTE: ConstructionTraits not implemented for SeqGS. See !303 in dune-istl
      else if (smoother == "gs" ||
               smoother == "gauss_seidel")
      {
        auto creator = typename Precon<Dune::SeqGS<Mat, Vec, Vec>>::Creator{};
        return creator.createWithString(prefix);
      }
#endif
#if 0
      // NOTE: ConstructionTraits not implemented for Richardson. See !304 in dune-istl
      else if (smoother == "richardson") {
        auto creator = typename Precon<Dune::Richardson<Vec, Vec>>::Creator{};
        return creator.createWithString(prefix);
      }
#endif
      else if (smoother == "ssor") {
        auto creator = typename Precon<Dune::SeqSSOR<Mat, Vec, Vec>>::Creator{};
        return creator.createWithString(prefix);
      }
      else {
        error_exit("Unknown smoother type '{}' given for parameter '{}'", smoother, prefix + "->smoother");
        return nullptr;
      }
    }
  };


  /// Implementation of \ref ISTLPreconCreatorInterface to be used in the
  /// \ref DefaultCreators.
  /**
   * Read several parameters of the AMG preconditioner and calls the corresponding
   * sub creators for the Smoother and the final AMG type.
   * 1. Depending on the solver category, create a LinearOperator,
   *    and (block) smoother, see \ref Dune::MatrixAdapter, \ref Dune::OverlappingSchwarzOperator,
   *    \ref Dune::NonoverlappingSchwarzOperator, \ref Dune::BlockPreconditioner, and
   *    \ref NonoverlappingBlockPreconditioner.
   * 2. Init the \ref Dune::Amg::SmootherTraits and the CoarsenCriterion depending on the
   *    Smoother type.
   * 3. Pass the created objects to \ref AMGCreator.
   *
   * Initfile parameters: (subset)
   * - `[PRECON]->preSmoothSteps:  Number of pre-smoother iterations [2]
   * - `[PRECON]->postSmoothSteps:  Number of post-smoother iterations [2]
   * - `[PRECON]->gamma: Number of two-grid cycles [1]
   * - `[PRECON]->smoother->...`: Parameters for the smoother
   * - `[PRECON]->coarsening->....`: Parameters for the Coarsening procedure
   * - `[PRECON]->aggregation->...`: Parameters for the coarsening by aggregation
   * - `[PRECON]->dependency->...`: Parameters for the characterization for variable dependencies.
   */
  template <class AMGCreator, class Smoother, class Traits>
  class AMGPrecon
      : public ISTLPreconCreatorInterface<Traits>
  {
    using Mat = typename Traits::Mat;
    using Vec = typename Traits::Vec;
    using Comm = typename Traits::Comm;

    using Super = ISTLPreconCreatorInterface<Traits>;
    using Self = AMGPrecon;

    using Interface = Dune::Preconditioner<Vec, Vec>;
    using LinOperator = typename Traits::LinOpCreator::Interface;

    using SolverCategory = typename Dune::SolverCategory::Category;

    static const bool is_arithmetic = std::is_arithmetic<typename Dune::FieldTraits<Mat>::field_type>::value;
    using Norm = std::conditional_t<is_arithmetic, Dune::Amg::FirstDiagonal, Dune::Amg::RowSum>;

    using SymCriterion = Dune::Amg::CoarsenCriterion<Dune::Amg::SymmetricCriterion<Mat,Norm>>;
    using UnSymCriterion = Dune::Amg::CoarsenCriterion<Dune::Amg::UnSymmetricCriterion<Mat,Norm>>;

  public:
    struct Creator : CreatorInterfaceName<Super>
    {
      std::unique_ptr<Super> createWithString(std::string prefix) override
      {
        return std::make_unique<Self>(prefix);
      }
    };

    AMGPrecon(std::string const& prefix)
      : prefix_(prefix)
    {
      initParameters(prefix);
    }

    /// Implements \ref ISTLPreconCreatorInterface::create
    std::unique_ptr<Interface> create(Mat const& mat, typename Traits::Comm const& comm) const override
    {
      return createImpl1(Dune::SolverCategory::category(comm),mat,comm.get());
    }

  private:
    template <class Communication,
      REQUIRES(not std::is_same<Communication, Dune::Amg::SequentialInformation>::value)>
    std::unique_ptr<Interface>
    createImpl1(SolverCategory cat, Mat const& mat, Communication const& comm) const
    {
      switch (cat) {
        case SolverCategory::sequential:
        {
          return createImpl1(cat, mat, Dune::Amg::SequentialInformation{});
        }
#if HAVE_MPI
        case SolverCategory::overlapping:
        {
          using LinOp = Dune::OverlappingSchwarzOperator<Mat,Vec,Vec,Communication>;
          using ParSmoother = Dune::BlockPreconditioner<Vec,Vec,Communication,Smoother>;
          LinOp* linOpPtr = new LinOp(mat, comm);
          linOperator_.reset(linOpPtr);
          return createImpl2<ParSmoother>(*linOpPtr, comm);
        }
        case SolverCategory::nonoverlapping:
        {
          using LinOp = Dune::NonoverlappingSchwarzOperator<Mat,Vec,Vec,Communication>;
          using ParSmoother = Dune::NonoverlappingBlockPreconditioner<Communication,Smoother>;
          LinOp* linOpPtr = new LinOp(mat, comm);
          linOperator_.reset(linOpPtr);
          return createImpl2<ParSmoother>(*linOpPtr, comm);
        }
#endif
        default:
          error_exit("Invalid solver category for AMGSolver\n");
          return nullptr; // avoid warnings
      }
    }

    std::unique_ptr<Interface>
    createImpl1(SolverCategory cat, Mat const& mat, Dune::Amg::SequentialInformation const& comm) const
    {
      DUNE_UNUSED_PARAMETER(cat);
      assert(cat == SolverCategory::sequential);
      using LinOp = Dune::MatrixAdapter<Mat,Vec,Vec>;
      LinOp* linOpPtr = new LinOp(mat);
      linOperator_.reset(linOpPtr);
      return createImpl2<Smoother>(*linOpPtr, comm);
    }

  private:
    template <class S, class LinOp, class Communication>
    std::unique_ptr<Interface> createImpl2(LinOp const& linOp, Communication const& comm) const
    {
      typename Dune::Amg::SmootherTraits<S>::Arguments smootherArgs;
      initSmootherParameters(prefix_ + "->smoother", smootherArgs);

      return symmetricAggregation_
        ? AMGCreator::template create<S>(prefix_, linOp, SymCriterion(parameters_), smootherArgs, comm)
        : AMGCreator::template create<S>(prefix_, linOp, UnSymCriterion(parameters_), smootherArgs, comm);
    }

  protected:
    void initParameters(std::string const& prefix)
    {
      // The debugging level. If 0 no debugging output will be generated.
      parameters_.setDebugLevel(Parameters::get<int>(prefix + "->debugLevel").value_or(2));
      // The number of presmoothing steps to apply
      parameters_.setNoPreSmoothSteps(Parameters::get<std::size_t>(prefix + "->preSmoothSteps").value_or(2));
      // The number of postsmoothing steps to apply
      parameters_.setNoPostSmoothSteps(Parameters::get<std::size_t>(prefix + "->postSmoothSteps").value_or(2));
      // Value of gamma; 1 for V-cycle, 2 for W-cycle
      parameters_.setGamma(Parameters::get<std::size_t>(prefix + "->gamma").value_or(1));
      // Whether to use additive multigrid.
      parameters_.setAdditive(Parameters::get<bool>(prefix + "->additive").value_or(false));
      // Whether to use symmetric or unsymmetric aggregation criterion
      symmetricAggregation_ = Parameters::get<bool>(prefix + "->symmetric aggregation").value_or(true);

      initCoarseningParameters(prefix + "->coarsening");
      initAggregationParameters(prefix + "->aggregation");
      initDependencyParameters(prefix + "->dependency");
    }

    void initCoarseningParameters(std::string const& prefix)
    {
      // The maximum number of levels allowed in the hierarchy.
      parameters_.setMaxLevel(Parameters::get<int>(prefix + "->maxLevel").value_or(100));
      // The maximum number of unknowns allowed on the coarsest level.
      parameters_.setCoarsenTarget(Parameters::get<int>(prefix + "->coarsenTarget").value_or(1000));
      // The minimum coarsening rate to be achieved.
      parameters_.setMinCoarsenRate(Parameters::get<double>(prefix + "->minCoarsenRate").value_or(1.2));
      // The damping factor to apply to the prolongated correction.
      parameters_.setProlongationDampingFactor(Parameters::get<double>(prefix + "->dampingFactor").value_or(1.6));
    }

    void initAggregationParameters(std::string const& prefix)
    {
      // Tthe maximal distance allowed between to nodes in a aggregate.
      parameters_.setMaxDistance(Parameters::get<std::size_t>(prefix + "->maxDistance").value_or(2));
      // The minimum number of nodes a aggregate has to consist of.
      parameters_.setMinAggregateSize(Parameters::get<std::size_t>(prefix + "->minAggregateSize").value_or(4));
      // The maximum number of nodes a aggregate is allowed to have.
      parameters_.setMaxAggregateSize(Parameters::get<std::size_t>(prefix + "->maxAggregateSize").value_or(6));
      // The maximum number of connections a aggregate is allowed to have.
      parameters_.setMaxConnectivity(Parameters::get<std::size_t>(prefix + "->maxConnectivity").value_or(15));
      // The maximum number of connections a aggregate is allowed to have.
      parameters_.setSkipIsolated(Parameters::get<bool>(prefix + "->skipIsolated").value_or(false));
    }

    void initDependencyParameters(std::string const& prefix)
    {
      // The scaling value for marking connections as strong.
      parameters_.setAlpha(Parameters::get<double>(prefix + "->alpha").value_or(1.0/3.0));
      // The threshold for marking nodes as isolated.
      parameters_.setBeta(Parameters::get<double>(prefix + "->beta").value_or(1.e-5));
    }

    template <class SA>
    void initSmootherParameters(std::string const& prefix, SA& smootherArgs) const
    {
      Parameters::get(prefix + "->iterations", smootherArgs.iterations);
      Parameters::get(prefix + "->relaxationFactor", smootherArgs.relaxationFactor);
    }

  private:
    std::string prefix_;
    Dune::Amg::Parameters parameters_;
    bool symmetricAggregation_ = true;

    mutable std::shared_ptr<LinOperator> linOperator_;
  };

} // end namespace AMDiS
