#pragma once

#include <dune/common/unused.hh>
#include <dune/istl/solver.hh>

#include <amdis/CreatorMap.hpp>
#include <amdis/Initfile.hpp>
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/linearalgebra/istl/CreatorInterfaces.hpp>

namespace AMDiS
{
  /// Implementation of \ref RunnerInterface for ISTL solvers
  template <class Traits>
  class ISTLRunner
      : public RunnerInterface<Traits>
  {
    using Mat = typename Traits::Mat;
    using Vec = typename Traits::Vec;

  public:
    explicit ISTLRunner(std::string const& prefix)
    {
      std::string solver = "default";
      Parameters::get(prefix, solver);

      using CreatorMap = CreatorMap<ISTLSolverCreatorInterface<Traits>>;
      auto* creator = named(CreatorMap::getCreator(solver, prefix));
      solverCreator_ = creator->createWithString(prefix);
      assert(solverCreator_);
    }

    /// Implements \ref RunnerInterface::init()
    void init(Mat const& mat, typename Traits::Comm& comm) override
    {
      solver_ = solverCreator_->create(mat, comm);
    }

    /// Implements \ref RunnerInterface::exit()
    void exit() override
    {
      solver_.reset();
    }

    /// Implements \ref RunnerInterface::solve()
    int solve(Mat const& A, Vec& x, Vec const& b, SolverInfo& solverInfo) override
    {
      DUNE_UNUSED_PARAMETER(A);

      // storing some statistics
      Dune::InverseOperatorResult statistics;

      // solve the linear system
      Vec _b = b;
      solver_->apply(x, _b, statistics);

      solverInfo.setRelResidual(statistics.reduction);

      return statistics.converged ? 0 : 1;
    }

  private:
    std::shared_ptr<ISTLSolverCreatorInterface<Traits>> solverCreator_;
    std::shared_ptr<typename Traits::Solver> solver_;
  };

} // end namespace AMDiS
