#pragma once

#include <list>
#include <string>
#include <memory>

#include <dune/istl/matrixindexset.hh>

#include <amdis/Output.hpp>
#include <amdis/linearalgebra/SymmetryStructure.hpp>

namespace AMDiS
{
  template <class T>
  class MatrixBackend
  {
  public:
    using Traits = T;

    /// The matrix type of the underlying base matrix
    using BaseMatrix = typename Traits::Mat;

    /// The type of the elements of the DOFMatrix
    using value_type = typename BaseMatrix::block_type;

    /// The index/size - type
    using size_type = typename BaseMatrix::size_type;

  public:
    /// Constructor. Constructs new BaseVector.
    template <class Basis>
    explicit MatrixBackend(Basis const&, Basis const&) {}

    /// Return the data-vector \ref vector
    BaseMatrix const& matrix() const
    {
      return matrix_;
    }

    /// Return the data-vector \ref vector
    BaseMatrix& matrix()
    {
      return matrix_;
    }

    /// create occupation pattern and apply it to the matrix
    template <class RowBasis, class ColBasis>
    void init(RowBasis const& rowBasis, ColBasis const& colBasis, SymmetryStructure symmetry)
    {
      auto occupationPattern = Dune::MatrixIndexSet{rowBasis.dimension(), colBasis.dimension()};

      auto rowLocalView = rowBasis.localView();
      auto colLocalView = colBasis.localView();
      for (const auto& element : elements(rowBasis.gridView())) {
        rowLocalView.bind(element);
        colLocalView.bind(element);

        for (std::size_t i = 0; i < rowLocalView.size(); ++i) {
          // The global index of the i-th vertex of the element
          auto row = rowLocalView.index(i);
          for (std::size_t j = 0; j < colLocalView.size(); ++j) {
            // The global index of the j-th vertex of the element
            auto col = colLocalView.index(j);
            occupationPattern.add(row, col);
          }
        }
      }

      occupationPattern.exportIdx(matrix_);

      symmetry_ = symmetry;
      initialized_ = true;
    }

    void finish()
    {
      initialized_ = false;
    }


    /// Insert a single value into the matrix (add to existing value)
    void insert(size_type r, size_type c, value_type const& value)
    {
      test_exit_dbg( initialized_, "Occupation pattern not initialized!");
      test_exit_dbg( r < matrix_.N() && c < matrix_.M() ,
          "Indices out of range [0,{})x[0,{})", matrix_.N(), matrix_.M() );
      matrix_[r][c] += value;
    }

    template <class Ind, class LocalMat>
    void scatter(Ind const& idx, LocalMat const& mat)
    {
      scatter(idx, idx, mat);
    }

    template <class RowInd, class ColInd, class LocalMat>
    void scatter(RowInd const& rows, ColInd const& cols, LocalMat const& mat)
    {
      test_exit_dbg( initialized_, "Occupation pattern not initialized!");
      for (size_type i = 0; i < size_type(rows.size()); ++i)
        for (size_type j = 0; j < size_type(cols.size()); ++j)
          matrix_[rows[i]][cols[j]] += mat[i][j];
    }


    SymmetryStructure symmetry() const
    {
      return symmetry_;
    }

    std::size_t nnz() const
    {
      return matrix_.nonzeroes();
    }

  private:
    BaseMatrix matrix_;

    bool initialized_ = false;
    SymmetryStructure symmetry_;
  };

} // end namespace AMDiS
