#pragma once

#include <amdis/CreatorInterface.hpp>
#include <amdis/CreatorMap.hpp>
#include <amdis/linearalgebra/istl/AMGPrecon.hpp>
#include <amdis/linearalgebra/istl/CreatorInterfaces.hpp>
#include <amdis/linearalgebra/istl/PreconWrapper.hpp>
#include <amdis/linearalgebra/istl/precompiled/Preconditioners.hpp>

namespace AMDiS
{
  namespace tag
  {
    struct bjacobi {};
    struct solver {};
  }

  /// Base class for precon creators, \see PreconCreator.
  /**
   * Constructor for preconditioners.
   *
   * Initfile parameters:
   * - `[PRECON]->relaxation`:  Dumping/relaxation factor
   * - `[PRECON]->iterations`:  Number of iterations the precon is applied.
   **/
  template <class Model, class Traits>
  struct ISTLPreconCreator
      : public ISTLPreconCreatorInterface<Traits>
  {
    using Interface = ISTLPreconCreatorInterface<Traits>;
    struct Creator : CreatorInterfaceName<Interface>
    {
      std::unique_ptr<Interface> createWithString(std::string prefix) override
      {
        return std::make_unique<Model>(prefix);
      }
    };

    explicit ISTLPreconCreator(std::string const& prefix)
    {
      Parameters::get(prefix + "->relaxation", w_);
      Parameters::get(prefix + "->iterations", iter_);
    }

  protected:
    double w_ = 1.0;
    int iter_ = 1;
  };


  /// Default precon creator.
  /**
   * Constructs a preconditioner, using the constructor signature
   * `Precon(Mat const& matrix, int iterations, double relaxation)`
   **/
  template <class Precon, class Traits>
  struct PreconCreator
      : public ISTLPreconCreator<PreconCreator<Precon,Traits>, Traits>
  {
    using Super = ISTLPreconCreator<PreconCreator, Traits>;
    using Super::Super; // inheriting constructor

    std::unique_ptr<typename Traits::Prec>
    create(typename Traits::Mat const& mat, typename Traits::Comm const& comm) const override
    {
      DUNE_UNUSED_PARAMETER(comm);
      return std::make_unique<Precon>(mat, this->iter_, this->w_);
    }
  };


  /// Precon creator for the Richardson preconditioner
  template <class Vec, class Traits>
  struct PreconCreator<Dune::Richardson<Vec, Vec>, Traits>
      : public ISTLPreconCreator<PreconCreator<Dune::Richardson<Vec, Vec>, Traits>, Traits>
  {
    using Super = ISTLPreconCreator<PreconCreator, Traits>;
    using Super::Super; // inheriting constructor

    std::unique_ptr<typename Traits::Prec>
    create(typename Traits::Mat const& mat, typename Traits::Comm const& comm) const override
    {
      DUNE_UNUSED_PARAMETER(mat);
      DUNE_UNUSED_PARAMETER(comm);
      using Precon = Dune::Richardson<Vec, Vec>;
      return std::make_unique<Precon>(this->w_);
    }
  };


  /// Precon creator for the SeqILDL preconditioner
  template <class Mat, class Vec, class Traits>
  struct PreconCreator<Dune::SeqILDL<Mat, Vec, Vec>, Traits>
      : public ISTLPreconCreator<PreconCreator<Dune::SeqILDL<Mat, Vec, Vec>, Traits>, Traits>
  {
    using Super = ISTLPreconCreator<PreconCreator, Traits>;
    using Super::Super; // inheriting constructor

    std::unique_ptr<typename Traits::Prec>
    create(typename Traits::Mat const& mat, typename Traits::Comm const& comm) const override
    {
      DUNE_UNUSED_PARAMETER(comm);
      using Precon = Dune::SeqILDL<Mat, Vec, Vec>;
      return std::make_unique<Precon>(mat, this->w_);
    }
  };


  /// Precon creator for the ParSSOR preconditioner.
  /**
   * Constructs a parallel SSOR preconditioner that can be used with
   * solverCategory == overlapping only.
   **/
  template <class Mat, class Vec, class Communication, class Traits>
  struct PreconCreator<Dune::ParSSOR<Mat, Vec, Vec, Communication>, Traits>
      : public ISTLPreconCreator<PreconCreator<Dune::ParSSOR<Mat, Vec, Vec, Communication>, Traits>, Traits>
  {
    using Super = ISTLPreconCreator<PreconCreator, Traits>;
    using Super::Super; // inheriting constructor

    std::unique_ptr<typename Traits::Prec>
    create(typename Traits::Mat const& mat, typename Traits::Comm const& comm) const override
    {
      test_exit(Dune::SolverCategory::category(comm) == Dune::SolverCategory::overlapping,
        "Dune::ParSSOR preconditioner can be used with overlapping domain decomposition.");

      using Precon = Dune::ParSSOR<Mat, Vec, Vec, Communication>;
      return std::make_unique<Precon>(mat, this->iter_, this->w_, comm.get());
    }
  };


  /// Precon creator for the InverseOperator2Preconditioner preconditioner.
  /**
   * Constructs a new solver that is wrapped into a preconditioner.
   *
   * Initfile parameters:
   * - `[PRECON]->solver`: the linear solver to use as preconditioner
   *
   * Note: The sub solver can be parametrized using the initfile parameters `[PRECON]->solver->(...)`.
   **/
  template <class Traits>
  struct PreconCreator<tag::solver, Traits>
      : public ISTLPreconCreator<PreconCreator<tag::solver, Traits>, Traits>
  {
    using Super = ISTLPreconCreator<PreconCreator, Traits>;

    explicit PreconCreator(std::string const& prefix)
      : Super(prefix)
    {
      std::string solver = "default";
      Parameters::get(prefix + "->solver", solver);

      using CreatorMap = CreatorMap<ISTLSolverCreatorInterface<Traits>>;
      auto* creator = named(CreatorMap::getCreator(solver, prefix + "->solver"));
      solverCreator_ = creator->createWithString(prefix + "->solver");
      assert(solverCreator_);
    }

    std::unique_ptr<typename Traits::Prec>
    create(typename Traits::Mat const& mat, typename Traits::Comm const& comm) const override
    {
      using InverseOp = Dune::InverseOperator<typename Traits::Vec, typename Traits::Vec>;
      using Precon = Dune::InverseOperator2Preconditioner<InverseOp>;
      using Wrapper = PreconWrapper<Precon, InverseOp>;
      return std::make_unique<Wrapper>(solverCreator_->create(mat, comm));
    }

  private:
    std::shared_ptr<ISTLSolverCreatorInterface<Traits>> solverCreator_;
  };


  /// Precon creator for the BJacobi preconditioner
  /**
   * Constructs a Block-Jacobi preconditioner with a sub-preconditioner
   * applied in each block.
   *
   * Initfile parameters:
   * - `[PRECON]->sub precon`:  The preconditioner used in each block
   *
   * NOTE: The sub preconditioner is constructed with sequential communication.
   * NOTE: The sub preconditioner can be parametrized using the initfile
   * parameters `[PRECON]->sub precon->(...)`.
   **/
  template <class Traits>
  struct PreconCreator<tag::bjacobi, Traits>
      : public ISTLPreconCreator<PreconCreator<tag::bjacobi, Traits>, Traits>
  {
    using Super = ISTLPreconCreator<PreconCreator, Traits>;
    using SeqTraits = SeqBackendTraits<Traits>;

    explicit PreconCreator(std::string const& prefix)
      : Super(prefix)
    {
      std::string subPrecon = "default";
      Parameters::get(prefix + "->sub precon", subPrecon);

      using CreatorMap = CreatorMap<ISTLPreconCreatorInterface<SeqTraits>>;
      auto* creator = named(CreatorMap::getCreator(subPrecon, prefix + "->sub precon"));
      subPreconCreator_ = creator->createWithString(prefix + "->sub precon");
      assert(subPreconCreator_);
    }

    std::unique_ptr<typename Traits::Prec>
    create(typename Traits::Mat const& mat, typename Traits::Comm const& comm) const override
    {
      return Traits::ParPrecCreator::create(Dune::SolverCategory::category(comm),
        subPreconCreator_->create(mat, comm.sequential()),
        comm);
    }

  private:
    std::shared_ptr<ISTLPreconCreatorInterface<SeqTraits>> subPreconCreator_;
  };


  /// Adds default creators for preconditioners for ISTL.
  /**
   * Adds creators for istl preconditioners.
   * - *diag*, *jacobi*: Diagonal preconditioner (Default), \see Dune::SeqJac
   * - *gs*, *gauss_seidel**: Gauss-Seidel preconditioner, \see Dune::SeqGS
   * - *sor*: Successive Overrelaxation methods, \see Dune::SeqSOR
   * - *ssor*: Symmetric Successive Overrelaxation methods, \see Dune::SeqSSOR
   * - *pssor*: A parallel SSOR preconditioner (requires overlap), \see Dune::ParSSOR
   * - *richardson*: Richardson methods, \see Dune::Richardson
   * - *solver*: Turns an InverseOperator into a Preconditioner, \see Dune::InverseOperator2Preconditioner
   * - *bjacobi*: Block-Jacobi methods, \see Dune::BlockPreconditioner, \see Dune::NoverlappingBlockPreconditioner
   * - *ilu,ilu0*: Incomplete LU factorization, \see Dune::SeqILU
   * - *ildl*: Incomplete LDL factorization, \see Dune::SeqILDL
   * - *amg*,*fastamg*,*kamg*: Algebraic multigrid methods, \see Dune::Amg::AMG, \see Dune::Amg::FastAMG, \see Dune::Amg::KAMG
   **/
  template <class Traits>
  class DefaultCreators<ISTLPreconCreatorInterface<Traits>>
  {
    using Mat = typename Traits::Mat;
    using Vec = typename Traits::Vec;

    template <class Precon>
    using PreconCreator = typename AMDiS::PreconCreator<Precon, Traits>::Creator;

    using Map = CreatorMap<ISTLPreconCreatorInterface<Traits>>;
    using FTraits = Dune::FieldTraits<typename Mat::field_type>;

  public:
    static void init()
    {
      auto jacobi = new PreconCreator<Dune::SeqJac<Mat, Vec, Vec>>;
      Map::addCreator("diag", jacobi);
      Map::addCreator("jacobi", jacobi);
      Map::addCreator("default", jacobi);

      auto gs = new PreconCreator<Dune::SeqGS<Mat, Vec, Vec>>;
      Map::addCreator("gs", gs);
      Map::addCreator("gauss_seidel", gs);

      auto sor = new PreconCreator<Dune::SeqSOR<Mat, Vec, Vec>>;
      Map::addCreator("sor", sor);

      auto ssor = new PreconCreator<Dune::SeqSSOR<Mat, Vec, Vec>>;
      Map::addCreator("ssor", ssor);

      init_ilu(std::is_arithmetic<typename FTraits::field_type>{});
      init_amg(std::is_same<typename FTraits::real_type, double>{});

      auto richardson = new PreconCreator<Dune::Richardson<Vec, Vec>>;
      Map::addCreator("richardson", richardson);

      auto solver = new PreconCreator<tag::solver>;
      Map::addCreator("solver", solver);

      init_bjacobi(Types<TYPEOF(std::declval<typename Traits::Comm>().get())>{}, Dune::PriorityTag<10>{});
    }

    static void init_ilu(std::false_type)
    {
      warning("ILU preconditioners not created for the matrix with field_type = {}.",
        Dune::className<typename FTraits::field_type>());
    }

    static void init_ilu(std::true_type)
    {
      auto ilu = new PreconCreator<Dune::SeqILU<Mat, Vec, Vec>>;
      Map::addCreator("ilu", ilu);
      Map::addCreator("ilu0", ilu);

      auto ildl = new PreconCreator<Dune::SeqILDL<Mat, Vec, Vec>>;
      Map::addCreator("ildl", ildl);
    }

    static void init_amg(std::false_type)
    {
      warning("AMG preconditioners not created for the matrix with real_type = {}.",
        Dune::className<typename FTraits::real_type>());
    }

    static void init_amg(std::true_type)
    {
      auto amg = new AMGPreconCreator<Dune::Amg::AMG, Traits>;
      Map::addCreator("amg", amg);
      auto fastamg = new AMGPreconCreator<Dune::Amg::FastAMG, Traits>;
      Map::addCreator("fastamg", fastamg);
      auto kamg = new AMGPreconCreator<Dune::Amg::KAMG, Traits>;
      Map::addCreator("kamg", kamg);
    }

    static void init_bjacobi(Types<Dune::Amg::SequentialInformation>, Dune::PriorityTag<2>) {}

    template <class Communication>
    static void init_bjacobi(Types<Communication>, Dune::PriorityTag<1>)
    {
      auto pssor = new PreconCreator<Dune::ParSSOR<Mat, Vec, Vec, Communication>>;
      Map::addCreator("pssor", pssor);

      auto bjacobi = new PreconCreator<tag::bjacobi>;
      Map::addCreator("bjacobi", bjacobi);
    }
  };


  // extern template declarations:
  extern template class DefaultCreators< ISTLPreconCreatorInterface<
    BackendTraits<typename AMDiS::YaspGridBasis<2,1>::GlobalBasis>> >;
  extern template class DefaultCreators< ISTLPreconCreatorInterface<
    BackendTraits<typename AMDiS::YaspGridBasis<2,2>::GlobalBasis>> >;

} // end namespace AMDiS
