#pragma once

#include <string>

// MTL4 headers
#include <boost/numeric/itl/itl.hpp>
#include <boost/numeric/mtl/mtl.hpp>

// AMDiS headers
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>

#include <amdis/linearalgebra/mtl/ITL_Preconditioner.hpp>

namespace AMDiS
{
  /** \ingroup Solver
   *
   * \brief
   * Wrapper class for different MTL4 itl-solvers. These solvers
   * are parametrized by Matrix and Vector.
   **/
  template <class ITLSolver, class Traits>
  class KrylovRunner
      : public RunnerInterface<Traits>
  {
    using Super = RunnerInterface<Traits>;
    using Matrix = typename Traits::Mat;
    using Vector = typename Traits::Vec;
    using Comm = typename Traits::Comm;
    using PreconBase = PreconditionerInterface<Traits>;
    using FloatType = typename Traits::CoefficientType;

  public:
    /// Constructor.
    explicit KrylovRunner(std::string const& prefix)
      : itlSolver_(prefix)
      , prefix_(prefix)
    {
      Parameters::get(prefix_ + "->absolute tolerance", aTol_);
      Parameters::get(prefix_ + "->relative tolerance", rTol_);
      Parameters::get(prefix_ + "->max iteration", maxIter_);
      Parameters::get(prefix_ + "->print cycle", printCycle_);
    }

    /// Implementation of \ref RunnerInterface::init()
    void init(Matrix const& A, Comm& comm) override
    {
      DUNE_UNUSED_PARAMETER(comm);
      initPrecon(prefix_);
      P_->init(A);
    }

    /// Implementation of \ref RunnerInterface::exit()
    void exit() override
    {
      P_->exit();
    }

    /// Implementation of \ref RunnerInterface::solve()
    int solve(Matrix const& A, Vector& x, Vector const& b, SolverInfo& solverInfo) override
    {
      test_exit(bool(P_), "There is no preconditioner");

      // print information about the solution process
      itl::cyclic_iteration<FloatType> iter(two_norm(b - A*x),
        maxIter_, rTol_, aTol_, printCycle_);
      iter.set_quite(solverInfo.info() == 0);
      iter.suppress_resume(solverInfo.info() == 0);

      int error = itlSolver_(A, x, b, *P_, iter);
      solverInfo.setAbsResidual(iter.resid());
      solverInfo.setRelResidual(iter.relresid());

      return error;
    }

  protected:
    /// Create left/right preconditioners from parameters given in the init-file
    void initPrecon(std::string const& prefix)
    {
      // Creator for the left preconditioner
      std::string preconName = "default";
      Parameters::get(prefix + "->precon", preconName);

      auto creator
        = named(CreatorMap<PreconBase>::getCreator(preconName, prefix + "->precon"));
      P_ = creator->createWithString(prefix + "->precon");
    }

  private:
    /// The itl solver object that performes the actual solve
    ITLSolver itlSolver_;

    /// The parameter prefix
    std::string prefix_;

    /// Pointer to the left/right preconditioner
    std::shared_ptr<PreconBase> P_;

    /// Absolute solver tolerance ||b - A*x|| < aTol
    FloatType aTol_ = 0;

    /// Relative solver tolerance ||b - A*x||/||b - A*x0|| < rTol
    FloatType rTol_ = 1.e-6;

    /// The maximal number of iterations
    std::size_t maxIter_ = 1000;

    /// Print solver residuals every \ref printCycle 's iteration
    std::size_t printCycle_ = 100;
  };

} // end namespace AMDiS
