#pragma once

#include <memory>

#include <amdis/CreatorMap.hpp>
#include <amdis/Initfile.hpp>
#include <amdis/Output.hpp>
#include <amdis/linearalgebra/LinearSolverInterface.hpp>
#include <amdis/linearalgebra/PreconditionerInterface.hpp>

namespace AMDiS
{
  /**
   * \ingroup Solver
   *
   * \brief Use a LinearSolver as Preconditioner
   */
  template <class Traits>
  class SolverPrecon
      : public PreconditionerInterface<Traits>
  {
    using Self = SolverPrecon;
    using Matrix = typename Traits::Mat;
    using Vector = typename Traits::Vec;
    using Super = PreconditionerInterface<Traits>;

  public:
    /// A creator to be used instead of the constructor.
    struct Creator : CreatorInterfaceName<Super>
    {
      std::unique_ptr<Super> createWithString(std::string prefix) override
      {
        return std::make_unique<Self>(std::move(prefix));
      }
    };

  public:
    SolverPrecon(std::string const& prefix)
      : solverInfo_(prefix + "->solver")
      , comm_(Environment::comm())
    {
      std::string solverName = "default";
      Parameters::get(prefix + "->solver", solverName);

      auto solverCreator = named(CreatorMap<LinearSolverInterface<Traits>>::getCreator(solverName, prefix + "->solver"));
      solver_ = solverCreator->createWithString(prefix + "->solver");
    }

    /// Implementation of \ref PreconditionerBase::init()
    void init(Matrix const& fullMatrix) override
    {
      matrix_ = &fullMatrix;
    }

    /// Implementation of \ref PreconditionerInterface::exit()
    void exit() override
    {
      matrix_ = nullptr;
    }

    /// Implementation of \ref PreconditionerBase::solve()
    void solve(Vector const& x, Vector& y) const override
    {
      test_exit_dbg(bool(solver_), "No solver initialized!");
      test_exit_dbg(bool(matrix_), "No matrix initialized!");

      y.checked_change_resource(x);
      test_exit(size(y) == num_cols(*matrix_), "incompatible size");
      solver_->solve(*matrix_, y, x, comm_, solverInfo_);
    }

    /// Implementation of \ref PreconditionerBase::adjoint_solve()
    void adjoint_solve(Vector const& x, Vector& y) const override
    {
      error_exit("Not Implemented.");
      test_exit_dbg(bool(solver_), "No solver initialized!");
      test_exit_dbg(bool(matrix_), "No matrix initialized!");

      y.checked_change_resource(x);
      test_exit(size(y) == num_rows(*matrix_), "incompatible size");
      // solver_->adjointSolve(*matrix_, y, x, comm_, solverInfo_);
    }

  private:
    mutable SolverInfo solverInfo_;
    mutable typename Traits::Comm comm_;

    std::shared_ptr<LinearSolverInterface<Traits>> solver_;
    Matrix const* matrix_ = nullptr;
  };

} // namespace AMDiS
