#pragma once

#include <algorithm>
#include <memory>
#include <vector>

#include <petscmat.h>

#include <amdis/Output.hpp>
#include <amdis/linearalgebra/SymmetryStructure.hpp>
#include <amdis/linearalgebra/petsc/MatrixNnzStructure.hpp>

namespace AMDiS
{
  /// \brief The basic container that stores a base matrix
  template <class T>
  class MatrixBackend
  {
    friend struct Constraints<MatrixBackend<T>>;

    using Communication = typename T::Comm;

  public:
    using Traits = T;

    /// The matrix type of the underlying base matrix
    using BaseMatrix = ::Mat;

    /// The type of the elements of the DOFMatrix
    using value_type = PetscScalar;

    /// The index/size - type
    using size_type = PetscInt;

  public:
    /// Constructor. Constructs new BaseMatrix.
    template <class Basis>
    explicit MatrixBackend(Basis const& rowBasis, Basis const& /*colBasis*/)
      : comm_(rowBasis.communicator())
    {}

    // disable copy and move operations
    MatrixBackend(MatrixBackend const&) = delete;
    MatrixBackend(MatrixBackend&&) = delete;
    MatrixBackend& operator=(MatrixBackend const&) = delete;
    MatrixBackend& operator=(MatrixBackend&&) = delete;

    ~MatrixBackend()
    {
      destroy();
    }

    /// Return a reference to the data-matrix \ref matrix
    BaseMatrix& matrix()
    {
      return matrix_;
    }

    /// Return a reference to the data-matrix \ref matrix
    BaseMatrix const& matrix() const
    {
      return matrix_;
    }

    /// Insert a single value into the matrix
    template <class RowIndex, class ColIndex>
    void insert(RowIndex const& r, ColIndex const& c, PetscScalar value)
    {
      PetscInt row = dofMap().global(r);
      PetscInt col = dofMap().global(c);
      MatSetValue(matrix_, row, col, value, ADD_VALUES);
    }

    /// Insert an element-matrix with row-indices == col-indices
    template <class LocalInd, class LocalMatrix>
    void scatter(LocalInd const& localInd, LocalMatrix const& localMat)
    {
      thread_local std::vector<PetscInt> idx;
      idx.resize(localInd.size());

      // create a vector of global indices from the local indices using the local-to-global map
      std::transform(localInd.begin(), localInd.end(), idx.begin(),
        [&](auto const& mi) { return this->dofMap().global(mi); });

      MatSetValues(matrix_, idx.size(), idx.data(), idx.size(), idx.data(), localMat.data(), ADD_VALUES);
    }

    /// Insert an element-matrix
    template <class RowLocalInd, class ColLocalInd, class LocalMatrix>
    void scatter(RowLocalInd const& rowLocalInd, ColLocalInd const& colLocalInd, LocalMatrix const& localMat)
    {
      thread_local std::vector<PetscInt> ri;
      thread_local std::vector<PetscInt> ci;
      ri.resize(rowLocalInd.size());
      ci.resize(colLocalInd.size());

      // create vectors of global indices from the local indices using the local-to-global map
      std::transform(rowLocalInd.begin(), rowLocalInd.end(), ri.begin(),
        [&](auto const& mi) { return this->dofMap().global(mi); });
      std::transform(colLocalInd.begin(), colLocalInd.end(), ci.begin(),
        [&](auto const& mi) { return this->dofMap().global(mi); });

      MatSetValues(matrix_, ri.size(), ri.data(), ci.size(), ci.data(), localMat.data(), ADD_VALUES);
    }

    /// Create and initialize the matrix
    template <class RowBasis, class ColBasis>
    void init(RowBasis const& rowBasis, ColBasis const& colBasis, SymmetryStructure symmetry)
    {
      Dune::Timer t;

      // update the non-zeros structure of the matrix.
      pattern_.update(rowBasis);
      info(3, "  update pattern needed {} seconds", t.elapsed());
      t.reset();

      // destroy an old matrix if created before
      destroy();
      info(3, "  destroy old matrix needed {} seconds", t.elapsed());
      t.reset();

      // create a MATAIJ or MATSEQAIJ matrix with provided sparsity pattern
      MatCreateAIJ(comm(),
        dofMap().localSize(), dofMap().localSize(),
        dofMap().globalSize(), dofMap().globalSize(),
        0, pattern_.d_nnz().data(), 0, pattern_.o_nnz().data(), &matrix_);

      // keep sparsity pattern even if we delete a row / column with e.g. MatZeroRows
      MatSetOption(matrix_, MAT_KEEP_NONZERO_PATTERN, PETSC_TRUE);

      // set symmetriy properties of the matrix
      switch (symmetry) {
        case SymmetryStructure::spd:
          MatSetOption(matrix_, MAT_SPD, PETSC_TRUE);
          break;
        case SymmetryStructure::symmetric:
          MatSetOption(matrix_, MAT_SYMMETRIC, PETSC_TRUE);
          break;
        case SymmetryStructure::hermitian:
          MatSetOption(matrix_, MAT_HERMITIAN, PETSC_TRUE);
          break;
        case SymmetryStructure::structurally_symmetric:
          MatSetOption(matrix_, MAT_STRUCTURALLY_SYMMETRIC, PETSC_TRUE);
          break;
        default:
          /* do nothing */
          break;
      }

      info(3, "  create new matrix needed {} seconds", t.elapsed());
      t.reset();

      initialized_ = true;
    }

    /// Finish assembly. Must be called before matrix can be used in a KSP
    void finish()
    {
      Dune::Timer t;
      MatAssemblyBegin(matrix_, MAT_FINAL_ASSEMBLY);
      MatAssemblyEnd(matrix_, MAT_FINAL_ASSEMBLY);
      info(3, "  finish matrix assembling needed {} seconds", t.elapsed());
    }

    /// Return the local number of nonzeros in the matrix
    std::size_t nnz() const
    {
      MatInfo info;
      MatGetInfo(matrix_, MAT_LOCAL, &info);
      return std::size_t(info.nz_used);
    }

  private:
    // The local-to-global map
    auto const& dofMap() const
    {
      return comm_.dofMap();
    }

    // An MPI Communicator or PETSC_COMM_SELF
    MPI_Comm comm() const
    {
#if HAVE_MPI
      return Environment::comm();
#else
      return PETSC_COMM_SELF;
#endif
    }

    // Destroy the matrix if initialized before.
    void destroy()
    {
      if (initialized_)
        MatDestroy(&matrix_);
    }

  private:
    Communication const& comm_;

    /// The data-matrix
    Mat matrix_;
    MatrixNnzStructure pattern_;

    bool initialized_ = false;
  };

} // end namespace AMDiS
