#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_divtrialvec {};
  }


  /// first-order operator \f$ \langle\psi, c\,\nabla\cdot\Phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::test_divtrialvec, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_divtrialvec, LC, GridFct>, LC, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( static_size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::test_divtrialvec, GridFct const& expr)
      : Super(expr, 1)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isPower,
        "row-node must be Leaf-Node and col-node must be a Power-Node.");

      static const std::size_t CHILDREN = CN::CHILDREN;
      static_assert( CHILDREN == CG::dow, "" );

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RangeFieldType = typename NodeQuadCache<typename CN::ChildType>::Traits::RangeFieldType;

      NodeQuadCache<RN> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename CN::ChildType> colCache(colLocalFE.localBasis());

      auto const& shapeValuesCache = rowCache.evaluateFunctionAtQP(contextGeo, quad);
      auto const& shapeGradientsCache = colCache.evaluateJacobianAtQP(contextGeo, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = contextGeo.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();

        // the values of the shape functions on the reference element at the quadrature point
        auto const& shapeValues = shapeValuesCache[iq];

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using WorldVector = FieldVector<RangeFieldType,CG::dow>;
        thread_local std::vector<WorldVector> colGradients;
        colGradients.resize(shapeGradients.size());

        for (std::size_t i = 0; i < colGradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], colGradients[i]);

        for (std::size_t i = 0; i < rowSize; ++i) {
          const auto local_i = rowNode.localIndex(i);
          const auto value = factor * shapeValues[i];
          for (std::size_t j = 0; j < colSize; ++j) {
            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_kj = colNode.child(k).localIndex(j);
              elementMatrix[local_i][local_kj] += value * colGradients[j][k];
            }
          }
        }
      }

    }
  };

  /** @} **/

} // end namespace AMDiS
