#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_trialvec {};
  }


  /// zero-order operator \f$ \langle\psi, \mathbf{b}\cdot\Phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::test_trialvec, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_trialvec, LC, GridFct>, LC, GridFct>
  {
    static const int dow = ContextGeometry<LC>::dow;
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( static_size_v<typename GridFct::Range> == dow, "Expression must be of vector type." );

  public:
    GridFunctionOperator(tag::test_trialvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isPower,
        "RN must be Leaf-Node and CN must be a Power-Node.");

      static const std::size_t CHILDREN = CN::CHILDREN;
      static_assert( static_size_v<typename GridFct::Range> == CHILDREN, "" );

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      NodeQuadCache<RN> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename CN::ChildType> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQP(contextGeo,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQP(contextGeo,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto b = Super::coefficient(local);

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowSize; ++i) {
          const auto local_i = rowNode.localIndex(i);

          for (std::size_t j = 0; j < colSize; ++j) {
            const auto value = b * (factor * rowShapeValues[i] * colShapeValues[j]);

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_kj = colNode.child(k).localIndex(j);
              elementMatrix[local_i][local_kj] += at(value,k);
            }
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
