#pragma once

#include <algorithm>
#include <type_traits>
#include <utility>
#include <vector>

#include <dune/typetree/typetree.hh>
#include <amdis/typetree/Traversal.hpp>

namespace AMDiS
{
  /**
  * \brief Container allowing to attach data to each node of a tree
  *
  * \ingroup Utility
  *
  * This provides operator[](Node) for accessing the data attached to the node.
  * For storing the data each node is identified via its treeIndex() method
  * which is supposed to return an index which is unique wrt the tree. These
  * indices need not to be consecutive but they are used to access an
  * internal vector<void*>. This may lead to wasted memory if the maximal
  * treeIndex() is much larger then the number of nodes within the tree.
  *
  * Before using the container it must be initialized by providing the
  * tree. The stored data objects will be created on initialization. Hence
  * the type of these data objects must be default constructible.
  *
  * Notice that the data per node can only be interpreted if the
  * node type is known. Hence the tree will be traversed on initilization,
  * copy, assignment, and destruction of a TreeData container.
  *
  * \tparam Basis Type of the basis with basis-tree
  * \tparam ND The data stored for a node of type Node will be of type ND<Node>
  * \tparam LO Set this flag if data should only be attached to leaf nodes.
  */
  template <class Basis, template<class> class ND, bool LO>
  class TreeData
  {
  public:
    //! Type of tree the data is associated with
    using Tree = typename Basis::LocalView::Tree;

    //! Type used for indices and size information
    using size_type = typename Tree::size_type;

    //! Set if data should only be associated to the leafs
    static const bool leafOnly = LO;

    //! Template to determine the data type for given node type
    template <class Node>
    using NodeData = ND<Node>;

  public:
    //! Default constructor
    TreeData()
      : basis_(nullptr)
    {}

    /**
      * \brief Construct from tree
      *
      * This default creates the data object associated to each node in the tree.
      * A reference to the tree is stored because it's needed for destruction
      * of the tree data.
      * See also \ref init.
      */
    explicit TreeData(Basis const& basis)
      : basis_(&basis)
    {
      initData();
    }

    //! Copy constructor
    TreeData(TreeData const& other)
      : basis_(other.basis_)
    {
      initData();
      copyData(other);
    }

    //! Move constructor
    TreeData(TreeData&& other)
      : TreeData()
    {
      swap(other);
    }

    /**
      * \brief Initialize from tree
      *
      * This default creates the data object associated to each node in the tree.
      * A reference to the tree is stored because it's needed for destruction
      * of the tree data.
      */
    void init(Basis const& basis)
    {
      destroyData();
      basis_ = &basis;
      initData();
    }

    //! Copy and Move assignment
    TreeData& operator=(TreeData other)
    {
      swap(other);
      return *this;
    }

    //! Destructor
    ~TreeData()
    {
      destroyData();
    }

    //! Get mutable reference to data associated to given node
    template<class Node>
    NodeData<Node>& operator[](const Node& node)
    {
      assert(initialized_);
      assert(data_.size() > node.treeIndex());
      using NodePtr = NodeData<Node>*;
      auto* nodePtr = NodePtr(data_[node.treeIndex()]);
      assert(nodePtr);
      return *nodePtr;
    }

    //! Get reference to data associated to given node
    template<class Node>
    const NodeData<Node>& operator[](const Node& node) const
    {
      assert(initialized_);
      assert(data_.size() > node.treeIndex());
      using NodePtr = NodeData<Node>*;
      auto* nodePtr = NodePtr(data_[node.treeIndex()]);
      assert(nodePtr);
      return *nodePtr;
    }

    //! Swap tree and data container with `other`
    void swap(TreeData& other)
    {
      using std::swap;
      swap(basis_, other.basis_);
      swap(data_, other.data_);
      swap(initialized_, other.initialized_);
    }

    Basis const* basis() const
    {
      return basis_;
    }

  protected:
    // For each node allocate memory and store the void* in the \ref data_
    void initData()
    {
      std::size_t s = 0;
      apply([&s](const auto& node, auto) -> void {
        s = std::max(s, node.treeIndex()+1);
      });

      data_.resize(s, nullptr);
      apply([this](const auto& node, auto) -> void {
        using Node = std::remove_reference_t<decltype(node)>;
        data_[node.treeIndex()] = new NodeData<Node>;
      });
      initialized_ = true;
    }

    // Deep copy of node data
    void copyData(const TreeData& other)
    {
      apply([&other,this](const auto& node, auto) -> void {
        (*this)[node] = other[node];
      });
    }

    // For each node, delete the allocated node data
    void destroyData()
    {
      apply([this](const auto& node, auto) -> void {
        using Node = std::remove_reference_t<decltype(node)>;
        using NodePtr = NodeData<Node>*;
        auto* p = NodePtr(data_[node.treeIndex()]);
        delete p;
      });
      basis_ = nullptr;
      initialized_ = false;
    }

  protected:
    // apply functor `func` to each (leaf) node of the tree, but only if `tree_` is set
    template <class Func>
    void apply(Func&& func)
    {
      if (basis_)
        applyImpl(func, std::integral_constant<bool, leafOnly>{});
    }

    template <class Func>
    void applyImpl(Func&& func, std::true_type) { for_each_leaf_node(basis_->localView().tree(), func); }

    template <class Func>
    void applyImpl(Func&& func, std::false_type) { for_each_node(basis_->localView().tree(), func); }

  protected:
    Basis const* basis_ = nullptr;
    std::vector<void*> data_;

    bool initialized_ = false;
  };


  namespace Impl
  {
    template <class Basis, template <class,class> class NodeData>
    struct RowNodeData
    {
      template <class RowNode>
      struct ColNodeData
      {
        template <class ColNode>
        using type = NodeData<RowNode, ColNode>;
      };

      template <class RowNode>
      using type = TreeData<Basis, ColNodeData<RowNode>::template type, false>;
    };

  } // end namespace Impl


  template <class RowBasis, class ColBasis, template <class,class> class NodeData,
    template <class> class ND = Impl::RowNodeData<ColBasis, NodeData>::template type>
  class MatrixData
      : public TreeData<RowBasis, ND, false>
  {
    using Super = TreeData<RowBasis, ND, false>;

  public:
    void init(RowBasis const& rowBasis, ColBasis const& colBasis)
    {
      Super::init(rowBasis);
      for_each_node(rowBasis.localView().tree(), [&](auto const& node, auto&&)
      {
        (*this)[node].init(colBasis);
      });
    }
  };


  template <class Basis, template <class> class NodeData>
  class VectorData
      : public TreeData<Basis, NodeData, false>
  {};

} // end namespace AMDiS
