#pragma once

#include <tuple>
#include <vector>

#include <dune/common/hash.hh>
#include <dune/geometry/quadraturerules.hh>
#include <dune/geometry/type.hh>

#include <amdis/common/ConcurrentCache.hpp>
#include <amdis/typetree/FiniteElementType.hpp>

namespace AMDiS
{
  /// \brief Cache of LocalBasis evaluations and jacobians at local coordinates
  /**
   * Caching is done using the ConcurrentCache data structure with a key that
   * depends on the element type and local coordinates inside the element.
   * Two methods are provided for evaluation of local basis functions and local
   * basis jacobians at the local coordinate.
   *
   * \tparam Node  A leaf basis node in a basis tree
   **/
  template <class Node>
  class NodeCache
  {
    static_assert(Node::isLeaf, "");
    using LocalBasisType = typename Node::FiniteElement::Traits::LocalBasisType;
    using DomainType = typename LocalBasisType::Traits::DomainType;

  public:
    using Traits = typename LocalBasisType::Traits;

    using ShapeValues = std::vector<typename Traits::RangeType>;
    using ShapeGradients = std::vector<typename Traits::JacobianType>;

  private:
    // Pair of GeometryType and local coordinates
    struct Key
    {
      unsigned int id;    // topologyId
      DomainType local;   // local coordinate

      struct hasher
      {
        std::size_t operator()(Key const& t) const
        {
          std::size_t seed = 0;
          Dune::hash_combine(seed, t.id);
          Dune::hash_range(seed, t.local.begin(), t.local.end());
          return seed;
        }
      };

      friend bool operator==(Key const& lhs, Key const& rhs)
      {
        return std::tie(lhs.id,lhs.local) == std::tie(rhs.id,rhs.local);
      }
    };

  public:
    /// Create a cache not associated to a local basis.
    /// Requires a call to \ref bind() before the first evaluation of the cache.
    NodeCache() = default;

    /// Associate the cache to a localBasis.
    NodeCache(LocalBasisType const& localBasis)
      : localBasis_(&localBasis)
    {}

    NodeCache(Node const& node)
      : NodeCache(node.finiteElement().localBasis())
    {}

    /// Bind the cache to a local basis for later evaluations.
    // NOTE: Should be called before the first evaluation of the local basis function or
    //       jacobian.
    // NOTE: Binding is only necessary, if the default constructor is used and
    //       the cache is not associated with a localBasis in the constructor.
    void bind(LocalBasisType const& localBasis)
    {
      localBasis_ = &localBasis;
    }

    /// Evaluate local basis functions at local coordinates
    ShapeValues const& evaluateFunction(Dune::GeometryType const& type, DomainType const& local) const
    {
      Key key{type.id(), local};
      return shapeValues_.get(key, [&](Key const&)
      {
        ShapeValues data;
        localBasis_->evaluateFunction(local, data);
        return data;
      });
    }

    /// Evaluate local basis jacobians at local coordinates
    ShapeGradients const& evaluateJacobian(Dune::GeometryType const& type, DomainType const& local) const
    {
      Key key{type.id(), local};
      return shapeGradients_.get(key, [&](Key const&)
      {
        ShapeGradients data;
        localBasis_->evaluateJacobian(local, data);
        return data;
      });
    }

  private:
    template <class Value>
    using Cache = ConcurrentCache<Key, Value, ThreadLocalPolicy,
      std::unordered_map<Key, Value, typename Key::hasher>>;

    Cache<ShapeValues> shapeValues_;
    Cache<ShapeGradients> shapeGradients_;

    LocalBasisType const* localBasis_ = nullptr;
  };

  /// Generator function for \ref NodeCache, \relates NodeCache.
  template <class Node>
  NodeCache<Node> makeNodeCache(Node const& node)
  {
    return NodeCache<Node>{node};
  }

  /// \brief Cache of LocalBasis evaluations and jacobians at quadrature points
  /**
   * Caching is done using the ConcurrentCache data structure with a key that
   * depends on the element type and order and size of a quadrature rule.
   * Two methods are provided for evaluation of local basis functions and local
   * basis jacobians at all quadrature points. A vector of values is returned.
   *
   * \tparam Node  A leaf basis node in a basis tree
   **/
  template <class Node>
  class NodeQuadCache
  {
    static_assert(Node::isLeaf, "");
    using LocalBasisType = typename Node::FiniteElement::Traits::LocalBasisType;

  public:
    using Traits = typename LocalBasisType::Traits;

    using ShapeValues = std::vector<typename Traits::RangeType>;
    using ShapeGradients = std::vector<typename Traits::JacobianType>;

    using ShapeValuesContainer = std::vector<ShapeValues>;
    using ShapeGradientsContainer = std::vector<ShapeGradients>;

  private:
    // Pair of GeometryType and quadrature order and size
    struct Key
    {
      unsigned int id;    // topologyId
      int p;              // quadrature order
      std::size_t size;   // number of quadrature points

      struct hasher
      {
        std::size_t operator()(Key const& t) const
        {
          std::size_t seed = 0;
          Dune::hash_combine(seed, t.id);
          Dune::hash_combine(seed, t.p);
          Dune::hash_combine(seed, t.size);
          return seed;
        }
      };

      friend bool operator==(Key const& lhs, Key const& rhs)
      {
        return std::tie(lhs.id,lhs.p,lhs.size) == std::tie(rhs.id,rhs.p,rhs.size);
      }
    };

  public:
    /// Create a cache not associated to a local basis.
    /// Requires a call to \ref bind() before the first evaluation of the cache.
    NodeQuadCache() = default;

    /// Associate the cache to a localBasis.
    NodeQuadCache(LocalBasisType const& localBasis)
      : localBasis_(&localBasis)
    {}

    NodeQuadCache(Node const& node)
      : NodeQuadCache(node.finiteElement().localBasis())
    {}

    /// Bind the cache to a local basis for later evaluations.
    // NOTE: Should be called before the first evaluation of the local basis function or
    //       jacobian.
    // NOTE: Binding is only necessary, if the default constructor is used and
    //       the cache is not associated with a localBasis in the constructor.
    void bind(LocalBasisType const& localBasis)
    {
      localBasis_ = &localBasis;
    }

    /// Evaluate local basis functions at all quadrature points of a quadrature rule.
    template <class CG, class QR>
    ShapeValuesContainer const& evaluateFunctionAtQP(CG const& contextGeo, QR const& quad) const
    {
      Key key{contextGeo.type().id(), quad.order(), quad.size()};
      return shapeValuesContainer_.get(key, [&](Key const&)
      {
        ShapeValuesContainer data(quad.size());
        for (std::size_t iq = 0; iq < quad.size(); ++iq)
          localBasis_->evaluateFunction(contextGeo.local(quad[iq].position()), data[iq]);
        return data;
      });
    }

    /// Evaluate local basis jacobians at all quadrature points of a quadrature rule.
    template <class CG, class QR>
    ShapeGradientsContainer const& evaluateJacobianAtQP(CG const& contextGeo, QR const& quad) const
    {
      Key key{contextGeo.type().id(), quad.order(), quad.size()};
      return shapeGradientsContainer_.get(key, [&](Key const&)
      {
        ShapeGradientsContainer data(quad.size());
        for (std::size_t iq = 0; iq < quad.size(); ++iq)
          localBasis_->evaluateJacobian(contextGeo.local(quad[iq].position()), data[iq]);
        return data;
      });
    }

  private:
    template <class Value>
    using Cache = ConcurrentCache<Key, Value, ThreadLocalPolicy,
      std::unordered_map<Key, Value, typename Key::hasher>>;

    Cache<ShapeValuesContainer> shapeValuesContainer_;
    Cache<ShapeGradientsContainer> shapeGradientsContainer_;

    LocalBasisType const* localBasis_ = nullptr;
  };

  /// Generator function for \ref NodeQuadCache, \relates NodeQuadCache.
  template <class Node>
  NodeQuadCache<Node> makeNodeQuadCache(Node const& node)
  {
    return NodeQuadCache<Node>{node};
  }

} // end namespace AMDiS
