#include "config.h"

#include <memory>

#include <dune/common/filledarray.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>

#include <amdis/AdaptiveGrid.hpp>
#include <amdis/AMDiS.hpp>
#include <amdis/LinearAlgebra.hpp>
#include <amdis/functions/ParallelGlobalBasis.hpp>

#include "Tests.hpp"

using namespace AMDiS;
using namespace Dune::Functions::BasisFactory;

template <class B, class T>
void test_dofvector(DOFVector<B,T>& vec)
{
  using size_type = typename DOFVector<B,T>::GlobalBasis::LocalView::size_type;
  using value_type = typename DOFVector<B,T>::value_type;

  auto const& basis = *vec.basis();
  AMDIS_TEST( vec.localSize() == basis.dimension() );
  vec.resizeZero();

  vec.init(false);
  vec.insert(0, value_type(1));
  vec.finish();

  value_type min(10);
  value_type max(-10);
  auto lv = basis.localView();
  auto const& gv = basis.gridView();
  for (auto const& e : elements(gv))
  {
    lv.bind(e);
    for (size_type i = 0; i < lv.size(); ++i)
    {
      auto value = vec.at(lv.index(i));
      min = std::min(min, value);
      max = std::max(max, value);
    }
  }
  AMDIS_TEST( min == T(0) || vec.localSize() == 1 );
  AMDIS_TEST( max == T(1) );
}

int main(int argc, char** argv)
{
  Environment env(argc, argv);

  // create grid
  using HostGrid = Dune::YaspGrid<2>;
  Dune::FieldVector<double, 2> L; L = 1.0;
  auto s = Dune::filledArray<2>(1);
  HostGrid hostGrid(L, s);
  AdaptiveGrid<HostGrid> grid(hostGrid);
  auto const& gridView = grid.leafGridView();

  // create basis
  auto preBasis = composite(power<2>(lagrange<2>(), flatInterleaved()),
                            lagrange<1>(),
                            flatLexicographic());
  auto basis = makeGlobalBasis(gridView, preBasis);

  using Basis = decltype(basis);

  {
    DOFVector<Basis> vec1(basis);
    test_dofvector(vec1);

    DOFVector<Basis, double> vec2(basis);
    test_dofvector(vec2);

    auto vec3 = makeDOFVector(basis);
    test_dofvector(vec3);

    auto vec4 = makeDOFVector<double>(basis);
    test_dofvector(vec4);

#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
    DOFVector vec5(basis);
    test_dofvector(vec5);
#endif
  }

  // test automatic updates
  {
    DOFVector<Basis> vec1(basis);

    // Conversion from Dune::Functions::DefaultGlobalBasis
    auto vec2 = makeDOFVector(makeBasis(gridView, preBasis));

    test_dofvector(vec1);
    for (auto const& e : elements(gridView))
      grid.mark(1, e);
    grid.preAdapt();
    grid.adapt();
    grid.postAdapt();
    std::size_t newSize = basis.dimension();
    AMDIS_TEST_EQ(vec1.localSize(), newSize);
    AMDIS_TEST_EQ(vec2.localSize(), newSize);
  }

  report_errors();
}
